#define INCL_DOS
#include <os2.h>
#include <malloc.h>

#define EANameString    "SOFTLINKTO"
#define EANameStringLen 10

char *__getSoftLinkTarget( HFILE fileHandle )
{
  EAOP2 eaList;
  FEA2LIST *feaListPtr;
  GEA2LIST *geaListPtr;
  int eaSize;
  ULONG rc;
  FILESTATUS4 fStats;
  char *linkTarget, *retPtr;
  
  rc = DosQueryFileInfo( fileHandle, FIL_QUERYEASIZE, &fStats,
   sizeof( FILESTATUS4 ) );
   
  if ( rc || !fStats.cbList )
    return NULL;
  
  eaSize = sizeof( GEA2LIST ) + EANameStringLen + 1;
  geaListPtr = (GEA2LIST *) malloc( eaSize );
  geaListPtr->cbList = eaSize;
  geaListPtr->list[0].oNextEntryOffset = 0;
  geaListPtr->list[0].cbName = EANameStringLen;
  strcpy( geaListPtr->list[0].szName, EANameString );

  eaSize = sizeof( FEA2LIST ) + fStats.cbList;
  feaListPtr = (FEA2LIST *) malloc( eaSize );
  feaListPtr->cbList = eaSize;

  eaList.fpFEA2List = feaListPtr;
  eaList.fpGEA2List = geaListPtr;
  
  rc = DosQueryFileInfo( fileHandle, FIL_QUERYEASFROMLIST, &eaList,
   sizeof( EAOP2 ) );
  
  if ( rc )
  {
    free( geaListPtr );
    free( feaListPtr );
    return NULL;
  }
  
  eaSize = strlen( feaListPtr->list[0].szName );
  
  if ( feaListPtr->list[0].cbValue && eaSize == EANameStringLen &&
        strcmp( EANameString, feaListPtr->list[0].szName ) == 0 )
  {
    linkTarget = feaListPtr->list[0].szName + feaListPtr->list[0].cbName + 1 +
     ( 2 * sizeof(USHORT) );
    if ( !strlen( linkTarget ) )
    {
      retPtr = NULL;
    } else {
      retPtr = (char *) malloc( strlen( linkTarget ) + 1 );
      strcpy( retPtr, linkTarget );
    }
  } else {
    retPtr = NULL;
  }
  
  free( geaListPtr );
  free( feaListPtr );
  
  return retPtr;
}

int readlink( char *srcPath, char *linkTargetBuffer, int targetBufferSize )
{
  ULONG rc, action;
  HFILE handle;
  char *myBuffer;
  int len;
  
  rc = DosOpen( srcPath, &handle, &action, 0, 0, OPEN_ACTION_FAIL_IF_NEW |
   OPEN_ACTION_OPEN_IF_EXISTS, OPEN_SHARE_DENYNONE, NULL );
  
  myBuffer = __getSoftLinkTarget( handle );
  DosClose( handle );
  
  if ( !myBuffer )
  {
    *linkTargetBuffer = 0;
    return -1;
  }
  
  len = strlen( myBuffer ) + 1;
  if ( len > targetBufferSize )
  {
    free( myBuffer );
    *linkTargetBuffer = 0;
    return -len;
  }
  
  strcpy( linkTargetBuffer, myBuffer );
  free( myBuffer );
  return len - 1;
}

int symlink( char *source, char *dest )
{
  HFILE fileHandle;
  EAOP2 eaList;
  FEA2LIST *feaListPtr;
  FEA2 *eaPtr;
  int eaSize;
  ULONG action, rc;
  
  eaList.fpGEA2List = NULL;
  eaSize = sizeof( FEA2LIST ) + strlen( EANameString ) + strlen( source ) + 2 +
   (2 * sizeof( USHORT ));
  feaListPtr = (FEA2LIST *) malloc( eaSize );
  eaList.fpFEA2List = feaListPtr;
  feaListPtr->cbList = eaSize;
  eaPtr = feaListPtr->list;
  
  eaPtr[0].oNextEntryOffset = 0;
  eaPtr[0].fEA = FEA_NEEDEA;
  eaPtr[0].cbName = strlen( EANameString );
  eaPtr[0].cbValue = strlen( source ) + (2 * sizeof( USHORT ));
  strcpy( eaPtr[0].szName, EANameString );
  ((USHORT *)(eaPtr[0].szName + strlen( EANameString ) + 1))[0] = 0xfffd;
  ((USHORT *)(eaPtr[0].szName + strlen( EANameString ) + 1))[1] =
   eaPtr[0].cbValue;
  strcpy( eaPtr[0].szName + eaPtr[0].cbName + 1 + (2 * sizeof( USHORT )),
   source );
  
  rc = DosOpen( dest, &fileHandle, &action, 0, FILE_NORMAL,
   OPEN_ACTION_CREATE_IF_NEW | OPEN_ACTION_REPLACE_IF_EXISTS,
   OPEN_SHARE_DENYNONE | OPEN_ACCESS_READWRITE, &eaList );
  
  free( feaListPtr );
  
  if ( rc )
  {
    return -1;
  }
  
  DosClose( fileHandle );
  return 0;
}

